# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Part of the high-level python bindings for Kopano

Copyright 2018 - 2019 Kopano and its licensors (see LICENSE file)
"""

import datetime
import re

import dateutil.parser

"""This module converts Microsoft KQL (Keyword Query Language)
queries into MAPI restrictions, as used when for example searching
for specific mails or users.

To achieve this, it uses a hand-written parser, using the building
blocks in parse.py.
"""

from MAPI import (
    FL_SUBSTRING, FL_IGNORECASE, RELOP_GT, RELOP_EQ, PT_BOOLEAN, PT_UNICODE,
    BMR_EQZ, BMR_NEZ, PT_SYSTIME, RELOP_GE, RELOP_LT, RELOP_LE, RELOP_NE,
    PT_SHORT, PT_LONG, PT_LONGLONG, PT_FLOAT, PT_DOUBLE, MNID_ID, MNID_STRING,
    PT_MV_UNICODE, MAPI_TO, MAPI_CC, MAPI_BCC
)

from MAPI.Tags import (
    PR_SUBJECT_W, PR_BODY_W, PR_MESSAGE_DELIVERY_TIME, PR_HASATTACH,
    PR_MESSAGE_SIZE, PR_MESSAGE_FLAGS, MSGFLAG_READ, PR_SENDER_NAME_W,
    PR_DISPLAY_NAME_W, PR_SENT_REPRESENTING_NAME_W, PR_SMTP_ADDRESS_W,
    PR_MESSAGE_ATTACHMENTS, PR_ATTACH_LONG_FILENAME_W, PR_MESSAGE_RECIPIENTS,
    PR_RECIPIENT_TYPE, PR_SENDER_EMAIL_ADDRESS_W, PR_EMAIL_ADDRESS_W,
    PR_GIVEN_NAME_W, PR_COMPANY_NAME_W, PR_SURNAME_W, PR_OFFICE_LOCATION_W,
    PR_MOBILE_TELEPHONE_NUMBER_W, PR_BUSINESS_TELEPHONE_NUMBER_W,
    PR_POSTAL_CODE_W, PR_BIRTHDAY, PR_PREFERRED_BY_NAME_W,
    PR_RESPONSIBILITY, PR_USER_NAME, PR_FOLDER_CHILD_COUNT,
    PR_CONTENT_COUNT, PR_PARENT_ENTRYID, PR_CONTENT_UNREAD
)
from MAPI.Struct import (
    SOrRestriction, SAndRestriction, SNotRestriction, SContentRestriction,
    SPropValue, SPropertyRestriction, SBitMaskRestriction, SSubRestriction,
)

from MAPI.Defs import PROP_TYPE

from .errors import ArgumentError
from .restriction import Restriction
from .defs import PSETID_Address, PS_PUBLIC_STRINGS
# Backwards compatible import for older python-mapi versions without datetime_to_filetime available in MAPI.Time
from . import utils as _utils

from .parse import (
    ParserInput, Parser, Char, CharSet, ZeroOrMore, OneOrMore, Sequence,
    Choice, Optional, Wrapper, NoMatch
)

# TODO such grouping: 'subject:(fresh exciting)'
# TODO Regex: avoid substr
# TODO OneOrMore(regex) not needed?
# TODO operator associativity/precedence (e.g. 'NOT a AND b'), check MSG
# TODO escaping double quotes
# TODO asterisk not implicit for phrases
# TODO relative dates rel. to timezone (e.g. "received:today")
# TODO graph does not support 'size>"10 KB" and such? we now roll our own
# TODO email matching on to/cc/bcc (PR_SEARCH_KEY/PR_EMAIL_ADDRESS?)
# TODO sender:user2@domain.com OR category:blue doesn't work, even if they
# separately work?

EMAIL1_NAME = (PSETID_Address, MNID_ID, 0x8083, PT_UNICODE) # TODO merge
CATEGORY_NAME = (PS_PUBLIC_STRINGS, MNID_STRING, 'Keywords', PT_MV_UNICODE)

RECIP_PROPS = [PR_DISPLAY_NAME_W, PR_EMAIL_ADDRESS_W, PR_SMTP_ADDRESS_W]

MESSAGE_KEYWORD_PROP = {
    'subject': PR_SUBJECT_W,
    'body': PR_BODY_W,
    'content': PR_BODY_W, # TODO what does content mean
    'received': PR_MESSAGE_DELIVERY_TIME,
    'hasAttachment': PR_HASATTACH,
    'hasAttachments': PR_HASATTACH,
    'size': PR_MESSAGE_SIZE,
    'read': (PR_MESSAGE_FLAGS, MSGFLAG_READ),
    'from': PR_SENT_REPRESENTING_NAME_W, # TODO email address
    # TODO why does 'from:user1@domain.com' work with just PR_SENDER_NAME_W!?
    'sender': [PR_SENDER_NAME_W, PR_SENDER_EMAIL_ADDRESS_W],
    'attachment': (PR_MESSAGE_ATTACHMENTS, PR_ATTACH_LONG_FILENAME_W),
    'category': CATEGORY_NAME,
    'to': (PR_MESSAGE_RECIPIENTS, RECIP_PROPS, MAPI_TO),
    'cc': (PR_MESSAGE_RECIPIENTS, RECIP_PROPS, MAPI_CC),
    'bcc': (PR_MESSAGE_RECIPIENTS, RECIP_PROPS, MAPI_BCC),
    'participants': (PR_MESSAGE_RECIPIENTS, RECIP_PROPS, None),
}

CONTACT_KEYWORD_PROP = {
    'name': PR_DISPLAY_NAME_W,
    'email': EMAIL1_NAME,
}

USER_KEYWORD_PROP = {
    'name': PR_DISPLAY_NAME_W,
    'givenName': PR_GIVEN_NAME_W,
    'preferredName': PR_PREFERRED_BY_NAME_W,
    'surname': PR_SURNAME_W,
    'userPrincipalName': PR_USER_NAME,
    'mail': PR_SMTP_ADDRESS_W,
    'mobilePhone': PR_MOBILE_TELEPHONE_NUMBER_W,
    'postalCode': PR_POSTAL_CODE_W,
    'birthday': PR_BIRTHDAY,
    'companyName': PR_COMPANY_NAME_W,
    'officeLocation': PR_OFFICE_LOCATION_W,
    'businessPhones': PR_BUSINESS_TELEPHONE_NUMBER_W,
    'responsibilities': PR_RESPONSIBILITY,
}

FOLDER_KEYWORD_PROP = {
    'displayName': PR_DISPLAY_NAME_W,
    'childFolderCount': PR_FOLDER_CHILD_COUNT,
    'parentFolderId': PR_PARENT_ENTRYID,
    'totalItemCount': PR_CONTENT_COUNT,
    'unreadItemCount': PR_CONTENT_UNREAD,
}

TYPE_KEYWORD_PROPMAP = {
    'message': MESSAGE_KEYWORD_PROP,
    'contact': CONTACT_KEYWORD_PROP,
    'user': USER_KEYWORD_PROP,
    'folder': FOLDER_KEYWORD_PROP,
}

DEFAULT_PROPTAGS = {
    'message': [PR_SUBJECT_W, PR_BODY_W, PR_SENT_REPRESENTING_NAME_W],
    'contact': [PR_DISPLAY_NAME_W, EMAIL1_NAME],
    'user': [PR_DISPLAY_NAME_W, PR_SMTP_ADDRESS_W],
    'folder': [PR_DISPLAY_NAME_W],
}

OP_RELOP = {
    '<': RELOP_LT,
    '>': RELOP_GT,
    '>=': RELOP_GE,
    '<=': RELOP_LE,
    '<>': RELOP_NE,
}


def _interval_restriction(proptag, start, end):
    start = _utils.datetime_to_filetime(start)
    end = _utils.datetime_to_filetime(end)

    return SAndRestriction([
        SPropertyRestriction(RELOP_GE, proptag, SPropValue(proptag, start)),
        SPropertyRestriction(RELOP_LT, proptag, SPropValue(proptag, end))
    ])

# AST node
class Term(object):
    def __init__(self, sign=None, field=None, op=None, value=None, hoepa=None):
        self.sign = sign
        self.field = field
        self.op = op
        self.value = value

    def restriction(self, type_, store):
        if self.field:
            # determine proptag for term, eg 'subject'
            proptag = TYPE_KEYWORD_PROPMAP[type_][self.field]
            flag = None
            subobj = None
            recipient_type = None

            # property in sub-object (attachments/recipient): sub-restriction
            if isinstance(proptag, tuple):
                if(proptag[0]) == PR_MESSAGE_ATTACHMENTS:
                    subobj, proptag = proptag
                elif(proptag[0]) == PR_MESSAGE_RECIPIENTS:
                    subobj, proptag, recipient_type = proptag
                elif len(proptag) == 2:
                    proptag, flag = proptag

                # named property: resolve local proptag
                elif len(proptag) == 4:
                    proptag = store._name_id(proptag[:3]) | proptag[3]

            # make restriction on proptag(s)
            if isinstance(proptag, list):
                restr = SOrRestriction([
                    self.prop_restriction(proptag, flag) for proptag in proptag
                ])
            else:
                restr = self.prop_restriction(proptag, flag)

            # turn restriction into sub-restriction
            if subobj:
                if recipient_type is not None:
                    restr = SAndRestriction([
                        restr,
                        SPropertyRestriction(
                                RELOP_EQ,
                                PR_RECIPIENT_TYPE,
                                SPropValue(PR_RECIPIENT_TYPE, recipient_type)
                        )
                    ])
                restr = SSubRestriction(subobj, restr)

        else:
            defaults = [(store._name_id(proptag[:3]) | proptag[3])
                           if isinstance(proptag, tuple) else proptag
                               for proptag in DEFAULT_PROPTAGS[type_]]

            restr = SOrRestriction([
                       SContentRestriction(
                           FL_SUBSTRING | FL_IGNORECASE,
                           p,
                           SPropValue(p, self.value)
                       ) for p in defaults
                   ])

        if self.sign == '-':
            restr = SNotRestriction(restr)

        return restr

    def prop_restriction(self, proptag, flag):
        # comparison operator
        if self.op in ('<', '>', '>=', '<=', '<>'):
            if PROP_TYPE(proptag) == PT_SYSTIME:
                d = dateutil.parser.parse(self.value)
                d = _utils.datetime_to_filetime(d)
                restr = SPropertyRestriction(
                            OP_RELOP[self.op],
                            proptag,
                            SPropValue(proptag, d)
                        )
            else:
                value = self.value
                unit = ''
                if [x for x in ('KB', 'MB', 'GB') if value.endswith(x)]:
                    value, unit = value[:-2], value[-2:]

                value = int(value)

                if unit == 'KB':
                    value *= 1024
                elif unit == 'MB':
                    value *= 1024**2
                elif unit == 'GB':
                    value *= 1024**3

                restr = SPropertyRestriction(
                            OP_RELOP[self.op],
                            proptag,
                            SPropValue(proptag, value)
                        )

        # contains/equals operator
        elif self.op in (':', '='):
            if PROP_TYPE(proptag) == PT_UNICODE:
                restr = SContentRestriction(
                            FL_SUBSTRING | FL_IGNORECASE,
                            proptag,
                            SPropValue(proptag, self.value)
                        )

            elif flag or PROP_TYPE(proptag) == PT_BOOLEAN:
                if flag:
                    restr = SBitMaskRestriction(
                        BMR_NEZ if self.value in ('yes', 'true') else BMR_EQZ,
                                proptag,
                                flag
                    )
                else:
                    restr = SPropertyRestriction(
                        RELOP_EQ,
                        proptag,
                        SPropValue(proptag, self.value in ('yes', 'true'))
                    )

            elif PROP_TYPE(proptag) == PT_MV_UNICODE:
                proptag2 = (proptag ^ PT_MV_UNICODE) | PT_UNICODE # funky!
                restr = SContentRestriction(
                            FL_SUBSTRING | FL_IGNORECASE,
                            proptag,
                            SPropValue(proptag2, self.value)
                        )

            elif (PROP_TYPE(proptag) in \
                  (PT_SHORT, PT_LONG, PT_LONGLONG, PT_FLOAT, PT_DOUBLE)):
                conv = float if PROP_TYPE(proptag) in (PT_FLOAT, PT_DOUBLE) \
                    else int
                if '..' in self.value:
                    val1, val2 = self.value.split('..')
                    restr = SAndRestriction([
                        SPropertyRestriction(
                            RELOP_GE,
                            proptag,
                            SPropValue(proptag, conv(val1))
                        ),
                        SPropertyRestriction(
                            RELOP_LT,
                            proptag,
                            SPropValue(proptag, conv(val2))
                        )
                    ])
                else:
                    restr = SPropertyRestriction(
                                RELOP_EQ,
                                proptag,
                                SPropValue(proptag, conv(self.value))
                            )

            elif PROP_TYPE(proptag) == PT_SYSTIME:
                if self.value == 'today':
                    d = datetime.datetime.now().date()
                    d2 = d + datetime.timedelta(days=1)
                    restr = _interval_restriction(proptag, d, d2)

                elif self.value == 'yesterday':
                    d2 = datetime.datetime.now().date()
                    d = d2 - datetime.timedelta(days=1)
                    restr = _interval_restriction(proptag, d, d2)

                elif self.value == 'this week':
                    d2 = datetime.datetime.now()
                    d = d2.date() - datetime.timedelta(days=d2.weekday())
                    restr = _interval_restriction(proptag, d, d2)

                elif self.value == 'this month':
                    d2 = datetime.datetime.now()
                    d = d2.date() - datetime.timedelta(days=d2.day-1)
                    restr = _interval_restriction(proptag, d, d2)

                elif self.value == 'last month':
                    now = datetime.datetime.now()
                    d2 = now.date() - datetime.timedelta(days=now.day-1)
                    d = (d2 - datetime.timedelta(days=1)).replace(day=1)
                    restr = _interval_restriction(proptag, d, d2)

                elif self.value == 'this year':
                    d2 = datetime.datetime.now()
                    d = datetime.datetime(d2.year, 1, 1)
                    restr = _interval_restriction(proptag, d, d2)

                elif self.value == 'last year':
                    now = datetime.datetime.now()
                    d2 = datetime.datetime(now.year, 1, 1)
                    d = datetime.datetime(d2.year-1, 1, 1)
                    restr = _interval_restriction(proptag, d, d2)

                elif '..' in self.value:
                    date1, date2 = self.value.split('..') # TODO hours etc
                    d = dateutil.parser.parse(date1)
                    d2 = dateutil.parser.parse(date2)
                    restr = _interval_restriction(proptag, d, d2)

                else:
                    d = dateutil.parser.parse(self.value) # TODO hours etc
                    d2 = d + datetime.timedelta(days=1)
                    restr = _interval_restriction(proptag, d, d2)

        return restr

class Operation(object):
    def __init__(self, op=None, args=None):
        self.op = op
        self.args = args

    def restriction(self, type_, store):
        if self.op == 'AND':
            return SAndRestriction(
                [arg.restriction(type_, store) for arg in self.args]
            )
        elif self.op == 'OR':
            return SOrRestriction(
                [arg.restriction(type_, store) for arg in self.args]
            )
        elif self.op == 'NOT':
            return SNotRestriction(
                self.args[0].restriction(type_, store)
            )

# build parser

class Regex(Parser):
    def __init__(self, regex):
        self._re = re.compile(regex)

    def parse(self, parser_input) :
        if parser_input.remaining() == 0:
            return NoMatch()
        rest = parser_input._data[parser_input._position:] # TODO slow
        match = self._re.match(rest)
        if not match:
            return NoMatch()
        else:
            n = match.end()
            parser_input.read(n)
            parser_input.inc_position(n)
            return self.match(rest[:n])

def _build_parser():
    whitespace = CharSet(' ')

    alphaspace = Regex(r'[\w \-+*@.]+')
    alphaplus = Regex(r'[\w\-+:<>=@.]+')

    word = OneOrMore(Regex(r'[\w+\-\*@\./]'))
    word.modifier = lambda t: ''.join(t)

    text = OneOrMore(alphaspace)
    text.modifier = lambda t: ''.join(t)

    quoted = Sequence(Char('"'), text, Char('"'))
    quoted.modifier = lambda t: ''.join(t)[1:-1]

    value = Choice(word, quoted)

    def op(s):
        operator = Sequence(*(Char(c) for c in s))
        operator.modifier = lambda t: ''.join(t)
        return operator

    operator = Choice(op(':'), op('='), op('<='), op('>='), op('<>'),
                      op('<'), op('>'))

    sign = CharSet('+-')

    term = Sequence(Optional(sign), Optional(Sequence(word, operator)), value)
    term.modifier = lambda t: Term(
        sign=t[0] if t[0] else None,
        field=t[1][0] if t[1] else None,
        op=t[1][1] if t[1] else None,
        value=t[2],
        hoepa=t
    )

    termplus = OneOrMore(alphaplus)
    termplus.modifier = lambda t: ''.join(t)

    term_fallback = Sequence(ZeroOrMore(whitespace), Choice(term, termplus))
    term_fallback.modifier = lambda t: \
        (Term(value=t[1]) if not isinstance(t[1], Term) else t[1])

    lpar = Sequence(ZeroOrMore(whitespace), op('('))
    rpar = Sequence(ZeroOrMore(whitespace), op(')'))

    expr = Wrapper()

    bracketed = Sequence(lpar, expr, rpar)
    bracketed.modifier = lambda t: t[1]

    unit = Choice(bracketed, term_fallback)

    and_ = Sequence(ZeroOrMore(whitespace), op('AND'))
    and_.modifier = lambda t: 'AND'
    or_ = Sequence(ZeroOrMore(whitespace), op('OR'))
    or_.modifier = lambda t: 'OR'
    not_ = Sequence(ZeroOrMore(whitespace), op('NOT'))
    not_.modifier = lambda t: 'NOT'

    wsexpr = Sequence(ZeroOrMore(whitespace), expr)
    wsexpr.modifier = lambda t: t[1]

    wsexpr2 = Sequence(OneOrMore(whitespace), expr)
    wsexpr2.modifier = lambda t: t[1]

    andor = Sequence(unit,
        Optional(Sequence(Optional(Choice(and_, or_)),
            Choice(bracketed, wsexpr))))
    def modifier(t):
        if t[1] is None:
            return t[0]
        else:
            return Operation(op=t[1][0] or 'AND', args=[t[0], t[1][1]])
    andor.modifier = modifier

    nott = Sequence(not_, Choice(bracketed, wsexpr2))
    nott.modifier = lambda t: Operation(op=t[0], args=[t[1]])

    expr.parser = Choice(nott, andor)

    return expr

_PARSER = _build_parser()

def _query_to_restriction(query, type_, store):
    query = str(query)
    try:
        ast = _PARSER.parse(ParserInput(query)).value
        return Restriction(ast.restriction(type_, store))
    except Exception:
        raise ArgumentError("could not process query")



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI import (
    MAPI_MODIFY, MAPI_DEFERRED_ERRORS
)

from MAPI.Tags import (
    PR_EC_HIERARCHYID, PR_ATTACH_NUM, PR_ATTACH_MIME_TAG_W, PR_RECORD_KEY,
    PR_ATTACH_LONG_FILENAME_W, PR_ATTACH_SIZE, PR_ATTACH_DATA_BIN, PR_ENTRYID,
    PR_LAST_MODIFICATION_TIME, IID_IAttachment, PR_ATTACH_METHOD,
    ATTACH_EMBEDDED_MSG, PR_ATTACH_DATA_OBJ, IID_IMessage,
    PR_ATTACHMENT_HIDDEN, PR_ATTACH_FLAGS, PR_ATTACH_CONTENT_ID_W,
    PR_ATTACH_CONTENT_LOCATION_W,
)
from MAPI.Defs import HrGetOneProp
from MAPI.Struct import MAPIErrorNoAccess

try:
    from . import item as _item
except ImportError: # pragma: no cover
    _item = sys.modules[__package__ + '.item']

from .errors import NotFoundError
from .properties import Properties

from .compat import (
    benc as _benc,
)

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']
try:
    from . import property_ as _prop
except ImportError: # pragma: no cover
    _prop = sys.modules[__package__ + '.property_']

class Attachment(Properties):
    """Attachment class

    Attachment abstraction. Attachments may contain binary data or
    embedded :class:`items <Item>`.

    Includes all functionality from :class:`Properties`.
    """

    def __init__(self, item, mapiitem=None, entryid=None, mapiobj=None):
        self._item = item
        self._mapiitem = mapiitem
        self._entryid = entryid
        self._mapiobj = mapiobj
        self._data = None

    @property
    def parent(self): # TODO would like to call this 'item' but already used
        """Parent :class:`item <Item>`."""
        return self._item

    @property
    def mapiobj(self):
        """Underlying MAPI object."""
        if self._mapiobj:
            return self._mapiobj

        self._mapiobj = self._mapiitem.OpenAttach(
            self._entryid, IID_IAttachment, 0
        )
        return self._mapiobj

    @mapiobj.setter
    def mapiobj(self, mapiobj):
        self._mapiobj = mapiobj

    def delete(self, objects):
        """Delete properties from attachment

        :param objects: The properties to delete
        """
        objs = _utils.arg_objects(
            objects,
            (_prop.Property,),
            'Attachment.delete'
        )
        proptags = [o.proptag for o in objs if isinstance(o, _prop.Property)]
        if proptags:
            self.mapiobj.DeleteProps(proptags)
        _utils._save(self._item.mapiobj)

    @property
    def entryid(self):
        return _benc(HrGetOneProp(self._mapiitem, PR_ENTRYID).Value) + \
               _benc(self[PR_RECORD_KEY])

    @property
    def hierarchyid(self):
        """Hierarchy (SQL) id."""
        return self.prop(PR_EC_HIERARCHYID).value

    @property
    def number(self):
        """Attachment number (respective to parent item)."""
        return self.get(PR_ATTACH_NUM, 0)

    @property
    def mimetype(self):
        """MIME type"""
        return self.get(PR_ATTACH_MIME_TAG_W, '')

    @mimetype.setter
    def mimetype(self, m):
        self[PR_ATTACH_MIME_TAG_W] = str(m)
        _utils._save(self._item.mapiobj)

    @property
    def filename(self):
        """Filename"""
        return self.get(PR_ATTACH_LONG_FILENAME_W, '')

    @property
    def hidden(self):
        """Is attachment hidden from end user."""
        return self.get(PR_ATTACHMENT_HIDDEN, False)

    @property
    def inline(self):
        """Is attachment inline."""
        return bool(self.get(PR_ATTACH_FLAGS, 0) & 4)

    @property
    def content_id(self):
        """Identifier used to reference (inline) attachment."""
        return self.get(PR_ATTACH_CONTENT_ID_W)

    @property
    def content_location(self):
        """URI pointing to contents of (inline) attachment."""
        return self.get(PR_ATTACH_CONTENT_LOCATION_W)

    @property
    def embedded(self):
        """Is attachment an embedded :class:`item <Item>`."""
        try:
            return self[PR_ATTACH_METHOD] == ATTACH_EMBEDDED_MSG
        except NotFoundError:
            return False

    @property
    def item(self):
        """Embedded :class:`item <Item>`."""
        try:
            msg = self.mapiobj.OpenProperty(PR_ATTACH_DATA_OBJ, IID_IMessage,
                0, MAPI_DEFERRED_ERRORS | MAPI_MODIFY)
        except MAPIErrorNoAccess:
            # TODO the following may fail for embedded items in certain
            # public stores, while the above does work (opening read-only
            # doesn't work, but read-write works! wut!?)
            msg = self.mapiobj.OpenProperty(PR_ATTACH_DATA_OBJ, IID_IMessage,
                0, MAPI_DEFERRED_ERRORS)
        item = _item.Item(mapiobj=msg)
        item.server = self._item.server
        return item

    @property
    def size(self):
        """Storage size (different from binary data size!)."""
        # TODO size of the attachment object, so more than just the attachment
        #      data
        # TODO (useful when calculating store size, for example.. sounds
        #      interesting to fix here)
        return self.get(PR_ATTACH_SIZE, 0)

    def __len__(self):
        return self.size

    @property
    def data(self):
        """Binary data."""
        if self._data is None:
            self._data = _utils.stream(self.mapiobj, PR_ATTACH_DATA_BIN)
        return self._data

    # file-like behaviour
    def read(self):
        return self.data

    @property
    def name(self):
        return self.filename

    @property
    def last_modified(self):
        """Last modification time."""
        return self.get(PR_LAST_MODIFICATION_TIME)

    def __unicode__(self):
        return 'Attachment("%s")' % (self.filename or '')



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import binascii

from MAPI import (
    MAPI_UNICODE, MAPI_UNRESOLVED, ECSTORE_TYPE_PRIVATE, ECSTORE_TYPE_ARCHIVE,
    WrapStoreEntryID
)
from MAPI.Defs import HrGetOneProp
from MAPI.Struct import (
    SPropValue, MAPIErrorNotFound, MAPIErrorInvalidParameter,
    MAPIErrorCollision, MAPIErrorNoSupport, ECUSER
)
from MAPI.Tags import (
    PR_EMAIL_ADDRESS_W, PR_DISPLAY_NAME_W, PR_EC_ENABLED_FEATURES_W,
    PR_EC_DISABLED_FEATURES_W, PR_EC_COMPANY_NAME_W,
    PR_MAPPING_SIGNATURE, PR_EC_ARCHIVE_SERVERS,
    EMS_AB_ADDRESS_LOOKUP, PR_GIVEN_NAME_W, PR_SURNAME_W,
    PR_MOBILE_TELEPHONE_NUMBER_W, PR_OFFICE_LOCATION_W,
    PR_TITLE_W, PR_EMS_AB_THUMBNAIL_PHOTO, PR_PREFERRED_BY_NAME_W,
    PR_BUSINESS_TELEPHONE_NUMBER_W, PR_BIRTHDAY, PR_POSTAL_CODE_W
)

from .store import Store
from .properties import Properties
from .group import Group
from .quota import Quota
from .defs import (
    ACTIVE_USER, NONACTIVE_USER,
)
from .errors import (
    Error, NotFoundError, NotSupportedError, DuplicateError, ArgumentError
)
from .compat import (
    benc as _benc, bdec as _bdec,
)
from .picture import Picture

from . import server as _server
from . import company as _company

class User(Properties):
    """User class.

    Includes all functionality from :class:`Properties`.

    :class:`Store` attributes can be accessed directly from a :class:`User` instance.
    """

    def __init__(self, name=None, server=None, email=None, ecuser=None,
                 userid=None):
        self.server = server or _server.Server(
            _skip_check=True, parse_args=False)
        self._ecuser = None
        self._name = None

        if ecuser:
            self._ecuser = ecuser

        elif userid:
            try:
                self._ecuser = \
                    self.server.sa.GetUser(_bdec(userid), MAPI_UNICODE)
            except (binascii.Error, TypeError): # TODO generalize?
                raise ArgumentError("invalid entryid: '%s'" % userid)
            except MAPIErrorNotFound:
                raise NotFoundError("no user found with userid '%s'" % userid)

        elif email or name:
            if email:
                try:
                    self._name = str(self.server.gab.ResolveNames(
                        [PR_EMAIL_ADDRESS_W],
                        MAPI_UNICODE | EMS_AB_ADDRESS_LOOKUP,
                        [[SPropValue(PR_DISPLAY_NAME_W, str(email))]],
                        [MAPI_UNRESOLVED]
                    )[0][0][1].Value)
                except (MAPIErrorNotFound, MAPIErrorInvalidParameter, \
                    IndexError):
                    raise NotFoundError("no such user '%s'" % email)
            else:
                self._name = str(name)

            try:
                self._ecuser = self.server.sa.GetUser(
                    self.server.sa.ResolveUserName(
                        self._name, MAPI_UNICODE), MAPI_UNICODE)
            # multi-tenant, but no '@' in username..
            except (MAPIErrorNotFound, MAPIErrorInvalidParameter):
                raise NotFoundError("no such user: '%s'" % self.name)

        if self._ecuser:
            self._name = self._ecuser.Username
            self._userid = _benc(self._ecuser.UserID)

        self._mapiobj = None

    @property
    def mapiobj(self):
        """Underlying MAPI object."""
        if not self._mapiobj:
            self._mapiobj = \
                self.server.mapisession.OpenEntry(self._ecuser.UserID, None, 0)
        return self._mapiobj

    @property
    def admin(self):
        """Is the user (regular or system) administrator."""
        return self._ecuser.IsAdmin >= 1

    @admin.setter
    def admin(self, value):
        self._update(admin=value)

    @property
    def admin_level(self):
        """User administration level (1 means regular administrator,
        2 means system administrator)."""
        return self._ecuser.IsAdmin

    @admin_level.setter
    def admin_level(self, value):
        self._update(admin=value)

    @property
    def hidden(self):
        """The user is hidden from the addressbook."""
        return bool(self._ecuser.IsHidden)

    @property
    def name(self):
        """User account name."""
        return self._name

    @name.setter
    def name(self, value):
        self._update(username=str(value))

    @property
    def fullname(self):
        """User full name."""
        return self._ecuser.FullName

    @fullname.setter
    def fullname(self, value):
        self._update(fullname=str(value))

    @property
    def email(self):
        """User email address."""
        return self._ecuser.Email

    @email.setter
    def email(self, value):
        self._update(email=str(value))

    @property
    def password(self):
        """User password (set-only)."""
        raise Error('passwords are write-only')

    @password.setter
    def password(self, value):
        self._update(password=str(value))

    @property
    def preferred_name(self):
        """User preferred name."""
        return self.get(PR_PREFERRED_BY_NAME_W, "")

    @property
    def business_phones(self):
        """User business phones.

        Note:
             Based on Microsoft Graph document:
             "Although this is a string collection, only one number
              can be set for this property."
        """
        phone = self.get(PR_BUSINESS_TELEPHONE_NUMBER_W)
        return [phone] if phone else []

    @property
    def birthday(self):
        """User birthday."""
        return self.get(PR_BIRTHDAY, "")

    @property
    def postal_code(self):
        """User postal code."""
        return self.get(PR_POSTAL_CODE_W, "")

    # TODO uniform with contact.photo.. class Picture (filename, dimensions..)?
    @property
    def photo(self):
        """User :class:`photo <Picture>`."""
        data = self.get(PR_EMS_AB_THUMBNAIL_PHOTO)
        if data is not None:
            return Picture(data=data)

    @property
    def features(self):
        """Enabled features (*pop3*, *imap*, *mobile*, ..)."""
        if not hasattr(self._ecuser, 'MVPropMap'):
            raise NotSupportedError('Python-Mapi does not support MVPropMap')

        for entry in self._ecuser.MVPropMap:
            if entry.ulPropId == PR_EC_ENABLED_FEATURES_W:
                return entry.Values

    @features.setter
    def features(self, value):
        if not hasattr(self._ecuser, 'MVPropMap'):
            raise NotSupportedError('Python-Mapi does not support MVPropMap')

        # Enabled + Disabled defines all features.
        features = \
            set([e for entry in self._ecuser.MVPropMap for e in entry.Values])
        disabled = list(features - set(value))

        # TODO: performance
        for entry in self._ecuser.MVPropMap:
            if entry.ulPropId == PR_EC_ENABLED_FEATURES_W:
                entry.Values = value
            if entry.ulPropId == PR_EC_DISABLED_FEATURES_W:
                entry.Values = disabled

        self._update()

    def add_feature(self, feature):
        """Enable a feature for a user.

        :param feature: The feature
        """
        feature = str(feature)
        if feature in self.features:
            raise DuplicateError("feature '%s' already enabled for user \
'%s'" % (feature, self.name))
        self.features = self.features + [feature]

    def remove_feature(self, feature):
        """Disable a feature for a user

        :param feature: The feature
        """
        # Copy features otherwise we will miss a disabled feature.
        # TODO: improvement?
        features = self.features[:]
        try:
            features.remove(str(feature))
        except ValueError:
            raise NotFoundError("no feature '%s' enabled for user '%s'" % \
                (feature, self.name))
        self.features = features
        self._update()

    @property
    def userid(self):
        """User id."""
        return self._userid

    @property
    def company(self):
        """:class:`Company` the user belongs to."""
        try:
            return _company.Company(HrGetOneProp(self.mapiobj,
                PR_EC_COMPANY_NAME_W).Value, self.server)
        except MAPIErrorNoSupport:
            return _company.Company('Default', self.server)

    @property
    def first_name(self):
        """User first name."""
        return self.get(PR_GIVEN_NAME_W, '')

    @property
    def job_title(self):
        """User job title."""
        return self.get(PR_TITLE_W, '')

    @property
    def mobile_phone(self):
        """User mobile phone number."""
        return self.get(PR_MOBILE_TELEPHONE_NUMBER_W, '')

    @property
    def office_location(self):
        """User office Location."""
        return self.get(PR_OFFICE_LOCATION_W, '')

    @property
    def last_name(self):
        """User last name."""
        return self.get(PR_SURNAME_W, '')

    @property # TODO
    def local(self):
        store = self.store
        return bool(store and (self.server.guid == \
            _benc(HrGetOneProp(store.mapiobj, PR_MAPPING_SIGNATURE).Value)))

    def create_store(self):
        """Create and attach store for user."""
        try:
            storeid_rootid = self.server.sa.CreateStore(
                ECSTORE_TYPE_PRIVATE, self._ecuser.UserID)
        except MAPIErrorCollision:
            raise DuplicateError("user '%s' already has store" % self.name)
        store_entryid = WrapStoreEntryID(0, b'zarafa6client.dll',
            storeid_rootid[0][:-4]) + self.server.pseudo_url + b'\x00'
        return Store(entryid=_benc(store_entryid), server=self.server)

    @property
    def store(self):
        """Return user :class:`Store` or *None* if no store is attached."""
        try:
            entryid = self.server.ems.CreateStoreEntryID(
                None, self._name, MAPI_UNICODE)
            return Store(entryid=_benc(entryid), server=self.server)
        except (MAPIErrorNotFound, NotFoundError):
            pass

    # TODO deprecated? user.store = .., user.archive_store = ..
    def hook(self, store):
        """Hook (attach) store.

        :param store: :class:`Store <Store>` to hook.
        """
        try:
            self.server.sa.HookStore(
                ECSTORE_TYPE_PRIVATE, _bdec(self.userid), _bdec(store.guid))
        except MAPIErrorCollision:
            raise DuplicateError("user '%s' already has hooked store" % \
                self.name)

    # TODO deprecated? user.store = None
    def unhook(self):
        """Unhook attached store."""
        try:
            self.server.sa.UnhookStore(
                ECSTORE_TYPE_PRIVATE, _bdec(self.userid))
        except MAPIErrorNotFound:
            raise NotFoundError("user '%s' has no hooked store" % self.name)

    def hook_archive(self, store):
        """Hook archive store.

        :param store: Archive :class:`store <Store>` to hook.
        """
        try:
            self.server.sa.HookStore(
                ECSTORE_TYPE_ARCHIVE, _bdec(self.userid), _bdec(store.guid))
        except MAPIErrorCollision:
            raise DuplicateError("user '%s' already has hooked archive \
store" % self.name)

    def unhook_archive(self):
        """Unhook archive store."""
        try:
            self.server.sa.UnhookStore(
                ECSTORE_TYPE_ARCHIVE, _bdec(self.userid))
        except MAPIErrorNotFound:
            raise NotFoundError("user '%s' has no hooked archive store" % \
                self.name)

    @property
    def active(self):
        """Is the user active."""
        return self._ecuser.Class == ACTIVE_USER

    @active.setter
    def active(self, value):
        if value:
            self._update(user_class=ACTIVE_USER)
        else:
            self._update(user_class=NONACTIVE_USER)

    @property
    def home_server(self):
        """Name of user home server."""
        return self._ecuser.Servername or self.server.name

    @property
    def archive_server(self):
        """Name of archive server."""
        try:
            return HrGetOneProp(self.mapiobj, PR_EC_ARCHIVE_SERVERS).Value[0]
        except MAPIErrorNotFound:
            return

    @property
    def quota(self):
        """User :class:`Quota`"""
        return Quota(self.server, self._ecuser.UserID)

    def groups(self):
        """Return all :class:`groups <Group>` to which user belongs."""
        for g in self.server.sa.GetGroupListOfUser(
                self._ecuser.UserID, MAPI_UNICODE):
            yield Group(g.Groupname, self.server)

    def send_as(self):
        """Return all :class:`users <User>` who can send-as this user."""
        for u in self.server.sa.GetSendAsList(
                self._ecuser.UserID, MAPI_UNICODE):
            yield self.server.user(u.Username)

    def add_send_as(self, user):
        """Add :class:`user <User>` as send-as.

        :param user: User to add.
        """
        try:
            self.server.sa.AddSendAsUser(
                self._ecuser.UserID, user._ecuser.UserID)
        except MAPIErrorCollision:
            raise DuplicateError("user '%s' already in send-as for user \
'%s'" % (user.name, self.name))

    def remove_send_as(self, user):
        """Remove :class:`user <User>` from send-as.

        :param user: User to remove.
        """
        try:
            self.server.sa.DelSendAsUser(
                self._ecuser.UserID, user._ecuser.UserID)
        except MAPIErrorNotFound:
            raise NotFoundError("no user '%s' in send-as for user '%s'" % \
                (user.name, self.name))

    def rules(self):
        """Return all :class:`rules <Rule>` on user inbox."""
        return self.inbox.rules()

    def __eq__(self, u): # TODO check same server?
        if isinstance(u, User):
            return self.userid == u.userid
        return False

    def __ne__(self, u):
        return not self == u

    def __unicode__(self):
        return "User(%s)" % (self._name or '')

    def _update(self, **kwargs):
        username = kwargs.get('username', self.name)
        password = kwargs.get('password', self._ecuser.Password)
        email = kwargs.get('email', str(self._ecuser.Email))
        fullname = kwargs.get('fullname', str(self._ecuser.FullName))
        user_class = kwargs.get('user_class', self._ecuser.Class)
        admin = kwargs.get('admin', self._ecuser.IsAdmin)

        try:
            # Pass the MVPropMAP otherwise the set values are reset
            if hasattr(self._ecuser, 'MVPropMap'):
                self.server.sa.SetUser(ECUSER(
                    Username=username, Password=password,
                    Email=email, FullName=fullname,
                    Class=user_class, UserID=self._ecuser.UserID,
                    IsAdmin=admin, MVPropMap=self._ecuser.MVPropMap
                ), MAPI_UNICODE)

            else:
                self.server.sa.SetUser(ECUSER(
                    Username=username, Password=password, Email=email,
                    FullName=fullname, Class=user_class,
                    UserID=self._ecuser.UserID, IsAdmin=admin
                ), MAPI_UNICODE)
        except MAPIErrorNoSupport:
            pass

        self._ecuser = self.server.sa.GetUser(
            self.server.sa.ResolveUserName(username, MAPI_UNICODE),
            MAPI_UNICODE
        )
        self._name = username

    def __getattr__(self, x):
        store = self.store
        if store:
            try:
                return getattr(store, x)
            except AttributeError:
                raise AttributeError("'User' object has no attribute '%s'" % x)

    def __setattr__(self, x, val):
        if x not in User.__dict__ and x in Store.__dict__:
            setattr(self.store, x, val)
        else:
            super(User, self).__setattr__(x, val)

    def __dir__(self):
        return list(User.__dict__) + list(Store.__dict__)



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI.Tags import (
    PR_FREEBUSY_ENTRYIDS, PR_PROCESS_MEETING_REQUESTS,
    PR_DECLINE_CONFLICTING_MEETING_REQUESTS,
    PR_DECLINE_RECURRING_MEETING_REQUESTS
)
from MAPI.Defs import HrGetOneProp
from MAPI.Struct import SPropValue
from MAPI import MAPI_MODIFY

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

class AutoAccept(object):
    """AutoAccept class

    Manage settings for automatically accepting meeting requests.
    """

    def __init__(self, store):
        fbeid = store.root.prop(PR_FREEBUSY_ENTRYIDS).value[1]
        self._fb = store.mapiobj.OpenEntry(fbeid, None, MAPI_MODIFY)

    @property
    def enabled(self):
        """Auto-accept is enabled."""
        return HrGetOneProp(self._fb, PR_PROCESS_MEETING_REQUESTS).Value

    @enabled.setter
    def enabled(self, b):
        self._fb.SetProps([SPropValue(PR_PROCESS_MEETING_REQUESTS, b)])
        _utils._save(self._fb)

    @property
    def conflicts(self):
        """Conflicting appointments are accepted."""
        prop = HrGetOneProp(self._fb, PR_DECLINE_CONFLICTING_MEETING_REQUESTS)
        return not prop.Value

    @conflicts.setter
    def conflicts(self, b):
        props = [SPropValue(PR_DECLINE_CONFLICTING_MEETING_REQUESTS, not b)]
        self._fb.SetProps(props)
        _utils._save(self._fb)

    @property
    def recurring(self):
        """Recurring appointments are accepted."""
        prop = HrGetOneProp(self._fb, PR_DECLINE_RECURRING_MEETING_REQUESTS)
        return not prop.Value

    @recurring.setter
    def recurring(self, b):
        props = [SPropValue(PR_DECLINE_RECURRING_MEETING_REQUESTS, not b)]
        self._fb.SetProps(props)
        _utils._save(self._fb)
